/*
 * Copyright (c) 1998,1999,2000,2001 Katsushi Kobayashi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the acknowledgement as bellow:
 *
 *      This product includes software developed by K. Kobayashi
 *
 *    And the paper of the driver code can be found at:
 *
 *    Katsushi Kobayashi, 
 *    "Design and Implementation of Firewire Device Driver on FreeBSD",
 *    Proc. of the FREENIX track: 1999 USENIX Annual Technical Coference,
 *    pp 41-51,(1999)
 *
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: sbp.c,v 1.8 2002/02/18 14:11:19 simokawa Exp $
 *
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <machine/bus.h>
#if __FreeBSD_version >= 500000
#include <machine/bus_dma.h>
#endif
#include <sys/malloc.h>
#include <sys/devicestat.h>	/* for struct devstat */


#include <cam/cam.h>
#include <cam/cam_ccb.h>
#include <cam/cam_sim.h>
#include <cam/cam_xpt_sim.h>
#include <cam/cam_debug.h>
#include <cam/cam_periph.h>

#include <cam/scsi/scsi_all.h>
#include <cam/scsi/scsi_message.h>
#include <cam/scsi/scsi_da.h>

#include <sys/kernel.h>

#include <vm/vm.h>
#include <vm/pmap.h>

#include <dev/firewire/firewire.h>
#include <dev/firewire/firewirereg.h>
#include <dev/firewire/iec13213.h>

#if 0
#define ccb_ocb_ptr	spriv_ptr0
#else
#define ccb_xfer_ptr	spriv_ptr0
#endif
#define ccb_sbp_ptr	spriv_ptr1

#define SBP_NUM_LUNS 64
#define SBP_INITIATOR 7
#define SBP_ESELECT_TIMEOUT 1

#define ORB_NOTIFY	(1 << 31)
#define	ORB_FMT_STD	(0 << 29)
#define	ORB_FMT_VED	(2 << 29)
#define	ORB_FMT_NOP	(3 << 29)
#define	ORB_FMT_MSK	(3 << 29)
#define	ORB_EXV		(1 << 28)
/* */
#define	ORB_CMD_IN	(1 << 27)
/* */
#define	ORB_CMD_SPD(x)	((x) << 24)
#define	ORB_CMD_MAXP(x)	((x) << 20)
#define	ORB_RCN_TMO(x)	((x) << 20)
#define	ORB_CMD_PTBL	(1 << 19)
#define	ORB_CMD_PSZ(x)	((x) << 16)

#define	ORB_FUN_LGI	(0 << 16)
#define	ORB_FUN_QLG	(1 << 16)
#define	ORB_FUN_RCN	(3 << 16)
#define	ORB_FUN_LGO	(7 << 16)
#define	ORB_FUN_MSK	(7 << 16)

#define SBP_DEBUG 0
#define VMACCESS 0

struct ind_ptr {
	u_int32_t hi,lo;
};
#define SBP_IND_MAX 0x20
struct sbp_ocb {
	struct sbp_ocb  *ocb;
	union ccb	*ccb;
	u_int32_t	orb[20];
	struct ind_ptr  ind_ptr[SBP_IND_MAX];
	u_int8_t	flags;
	u_int8_t	tag;
	u_int8_t	target;
	u_int8_t	lun;
	u_int8_t	status;
	u_int8_t	cmd_len;
	u_int8_t	*cmd_ptr;
	u_int32_t	data_len;
	u_int8_t	*data_ptr;
};
#define OCB_ACT_MGM 0
#define OCB_ACT_CMD 1
#define OCB_ACT_MASK 3

#define SBP_RESOURCE_SHORTAGE 0x10

struct sbp_login_res{
#if FW_ENDIANSWAP == 0 && BYTE_ORDER == LITTLE_ENDIAN
	u_int16_t	len;
	u_int16_t	id;
	u_int16_t	res0;
	u_int16_t	cmd_hi;
	u_int32_t	cmd_lo;
	u_int16_t	res1;
	u_int16_t	recon_hold;
#else
	u_int16_t	id;
	u_int16_t	len;
	u_int16_t	cmd_hi;
	u_int16_t	res0;
	u_int32_t	cmd_lo;
	u_int16_t	recon_hold;
	u_int16_t	res1;
#endif
};
struct sbp_status{
#if FW_ENDIANSWAP == 0 && BYTE_ORDER == LITTLE_ENDIAN
	u_int8_t	len:3,
			dead:1,
			resp:2,
			src:2;
	u_int8_t	status:8;
	u_int16_t	orb_hi;
	u_int32_t	orb_lo;
	u_int32_t	data[6];
#else
	u_int16_t	orb_hi;
	u_int8_t	status:8;
	u_int8_t	len:3,
			dead:1,
			resp:2,
			src:2;
	u_int32_t	orb_lo;
	u_int32_t	data[6];
#endif
};
struct sbp_cmd_status{
#define SBP_SFMT_CURR 0
#define SBP_SFMT_DEFER 1
#if FW_ENDIANSWAP == 0 && BYTE_ORDER == LITTLE_ENDIAN
	u_int8_t	status:6,
			sfmt:2;
	u_int8_t	s_key:4,
			ill_len:1,
			eom:1,
			mark:1,
			valid:1;
	u_int8_t	s_code;
	u_int8_t	s_qlfr;
	u_int32_t	info;
	u_int32_t	cdb;
	u_int32_t	s_keydep:24,
			fru:8;
	u_int32_t	vend[2];
#else
	u_int8_t	s_qlfr;
	u_int8_t	s_code;
	u_int8_t	s_key:4,
			ill_len:1,
			eom:1,
			mark:1,
			valid:1;
	u_int8_t	status:6,
			sfmt:2;
	u_int32_t	info;
	u_int32_t	cdb;
	u_int32_t	s_keydep:24,
			fru:8;
	u_int32_t	vend[2];
#endif
};
struct sbp_softc {
	struct firewire_dev_comm fd;
	unsigned char flags;
	int transfer_dir;

	void *transfer_data;
	int transfer_datalen;
	int transfer_actlen;
	struct cam_sim  *sim;
/*
	transfer_cb_f transfer_cb;
*/
	void *transfer_priv;
	int transfer_status;
	int transfer_state;

	int sbp_stat;
	int sbp_count;
	int sbp_error;

	unsigned char cam_scsi_command[CAM_MAX_CDBLEN];
/*
	struct scsi_sence cam_scsi_sense;
*/
	struct sbp_dev{
		struct fw_device *fwdev;
#define SBP_DEV_RESET		0
#define SBP_DEV_RECONN		1
#define SBP_DEV_TOATTACH	3
#define SBP_DEV_ATTACHED	4
		int status;
		u_int32_t mgm_hi, mgm_lo;
		struct sbp_login_res login;
		struct sbp_ocb *ocbs;
	} dev[SBP_NUM_LUNS];
	struct fw_bind fwb;
	struct sbp_ocb *free_ocbs;
	struct sbp_ocb ocb[257];
	struct sbp_ocb *dummy_ocb;
	bus_dma_tag_t	dmat;
};

static void sbp_post_explore __P((void *));
static void sbp_do_attach __P((struct sbp_softc *));
static void sbp_recv __P((struct fw_xfer *));
static void sbp_login_callback __P((struct fw_xfer *));
static void sbp_login __P((struct sbp_softc *));
static void sbp_login_dev __P((struct sbp_softc *, int));
static void sbp_execute_ocb __P((void *,  bus_dma_segment_t *, int, int));
static void sbp_free_ocb __P((struct sbp_softc *, struct sbp_ocb *));
static struct sbp_ocb * sbp_get_ocb __P((struct sbp_softc *));
static void sbp_enqueue_ocb __P((struct sbp_softc *, int, struct sbp_ocb *));
static struct sbp_ocb * sbp_dequeue_ocb __P((struct sbp_softc *, int ));
static void sbp_timeout __P((void *arg));

#define SBPPRI ((PZERO+8)|PCATCH)

#if 0
#define SBP_DEV2LUN(sbp, x) (((x) - (sbp->dev)) / sizeof(struct sbp_dev))
#else
#define SBP_DEV2LUN(sbp, x) ((struct sbp_dev *)(x) - (sbp->dev))
#endif

/* cam related functions */
static void	sbp_action(struct cam_sim *sim, union ccb *ccb);
static void	sbp_poll(struct cam_sim *sim);
static void	sbp_cam_rescan_callback(struct cam_periph *periph,
					union ccb *ccb);
static void	sbp_cam_rescan(struct sbp_softc *sbp);

static void
sbp_identify(driver_t *driver, device_t parent)
{
	device_t child;
#if SBP_DEBUG
	printf("sbp_identify\n");
#endif

	BUS_ADD_CHILD(parent, 0, "sbp", 0);
	device_set_driver(child, driver);
}

/*
 * sbp_probe()
 */
static int
sbp_probe(device_t dev)
{
	struct sbp_softc *sbp;
	device_t pa;

#if SBP_DEBUG
	printf("sbp_probe\n");
#endif

	pa = device_get_parent(dev);
	if(device_get_unit(dev) != device_get_unit(pa)){
		return(ENXIO);
	}

	sbp = ((struct sbp_softc *)device_get_softc(dev));
	bzero(sbp, sizeof(struct sbp_softc));
	sbp->fd.dev = dev;
	sbp->fd.fc = device_get_ivars(dev);
	sbp->fd.post_explore = sbp_post_explore;
	device_set_desc(dev, "SBP2/SCSI over firewire");
	return (0);
}

void sbp_post_explore(void *arg)
{
	struct sbp_softc *sbp = (struct sbp_softc *)arg;
	struct fw_device *fwdev;
	int i, j;

#if SBP_DEBUG
	printf("sbp_post_explore\n");
#endif
	for(i = 0 ; i < SBP_NUM_LUNS ; i ++){
		for( fwdev  = STAILQ_FIRST(&sbp->fd.fc->devices);
			fwdev != NULL; fwdev = STAILQ_NEXT(fwdev, link)){
			if(sbp->dev[i].fwdev == NULL) break;
			if(sbp->dev[i].fwdev == fwdev) break;
		}
		if(fwdev == NULL){
			sbp->dev[i].fwdev = NULL;
			sbp->dev[i].login.id = 0;
			sbp->dev[i].login.cmd_hi = 0;
			sbp->dev[i].login.cmd_lo = 0;
			sbp->dev[i].ocbs = NULL;
		}
		if(sbp->dev[i].fwdev != NULL){
			sbp->dev[i].status = SBP_DEV_RECONN;
		}else{
			sbp->dev[i].status = SBP_DEV_RESET;
		}
	}
	for( fwdev  = STAILQ_FIRST(&sbp->fd.fc->devices);
		fwdev != NULL; fwdev = STAILQ_NEXT(fwdev, link)){
		if(fwdev->status != FWDEVATTACHED) break;
		if(!(getcsrdata(fwdev, CSRKEY_SPEC) == CSRVAL_ANSIT10
			 && getcsrdata(fwdev, CSRKEY_VER) == CSRVAL_T10SBP2)) break;
		for(i = 0 ; i < SBP_NUM_LUNS ; i ++){
			if(sbp->dev[i].fwdev == fwdev ) break;
		}
		if( i == SBP_NUM_LUNS){
			for( j = 0 ; j < SBP_NUM_LUNS ; j++)
				if(sbp->dev[j].fwdev == NULL) break;
			if( j == SBP_NUM_LUNS ) return;

#if SBP_DEBUG
			printf("SBP-II: Found target %d: node=%d\n", j, fwdev->dst);
#endif
			sbp->dev[j].fwdev = fwdev;
			sbp->dev[j].login.cmd_hi = 0;
			sbp->dev[j].login.cmd_lo = 0;
			if((sbp->dev[j].mgm_lo = getcsrdata(fwdev, 0x54))
				== 0 ){
				return;
			}
			sbp->dev[j].mgm_hi = 0xffff;
			sbp->dev[j].mgm_lo = 0xf0000000 | sbp->dev[j].mgm_lo << 2;
		}
	}
	sbp_login(sbp);
}
static void sbp_loginres_callback(struct fw_xfer *xfer){
#if SBP_DEBUG
	printf("sbp_loginres_callback\n");
#endif
	fw_xfer_free(xfer);
	return;
}
static void sbp_login_callback(struct fw_xfer *xfer){
#if SBP_DEBUG
	printf("sbp_login_callback\n");
#endif
	fw_xfer_free(xfer);
	return;
}
static void sbp_bto_callback(struct fw_xfer *xfer)
{
	struct sbp_softc *sbp;
#if SBP_DEBUG
	printf("sbp_bto_callback\n");
#endif
	sbp = (struct sbp_softc *)xfer->sc;
	fw_xfer_free(xfer);
	sbp_do_attach(sbp);
	sbp_login(sbp);
	return;
}
static void sbp_do_attach(struct sbp_softc *sbp){
	int i;
#if SBP_DEBUG
	printf("sbp_do_attach\n");
#endif
	for( i = 0 ; i < SBP_NUM_LUNS ; i++){
		if((sbp->dev[i].fwdev != NULL) &&
		   (sbp->dev[i].fwdev->status == FWDEVATTACHED) &&
		   (sbp->dev[i].status == SBP_DEV_TOATTACH)){
			break;
		}
	}
	sbp->dev[i].status = SBP_DEV_ATTACHED;
}
static void sbp_bto(struct sbp_softc *sbp)
{
	struct fw_xfer *xfer;
	struct fw_pkt *fp;
	int i;

#if SBP_DEBUG
	printf("sbp_bto\n");
#endif
	for( i = 0 ; i < SBP_NUM_LUNS ; i++){
		if((sbp->dev[i].fwdev != NULL) &&
		   (sbp->dev[i].fwdev->status == FWDEVATTACHED) &&
		   (sbp->dev[i].status == SBP_DEV_TOATTACH)){
			break;
		}
	}
	if(i < SBP_NUM_LUNS){
		xfer = fw_xfer_alloc();
		if(xfer == NULL){
			return;
		}
		xfer->send.len = 16;
		xfer->send.buf = malloc(xfer->send.len, M_DEVBUF, M_NOWAIT);
		if(xfer->send.buf == NULL){
			fw_xfer_free( xfer);
			return;
		}
		xfer->send.off = 0; 
		xfer->spd = 0;
		xfer->sc = (caddr_t)sbp;
		xfer->fc = sbp->fd.fc;
		xfer->act.hand = sbp_bto_callback;
		xfer->retry_req = fw_asybusy;

		fp = (struct fw_pkt *)xfer->send.buf;
		fp->mode.wreqq.dest_hi = htons(0xffff);
		fp->mode.wreqq.dest_lo = htonl(0xf0000210);
		fp->mode.wreqq.tlrt = 0;
		fp->mode.wreqq.tcode = FWTCODE_WREQQ;
		fp->mode.wreqq.pri = 0;
		xfer->dst = 0xffc0 | sbp->dev[i].fwdev->dst;
		fp->mode.wreqq.dst = htons(xfer->dst);
		fp->mode.wreqq.data = htonl(0xf);
		fw_asyreq(xfer->fc, -1, xfer);
	}
}
static void sbp_resetdev_callback(struct fw_xfer *xfer)
{
	struct sbp_softc *sbp;
#if SBP_DEBUG
	printf("sbp_resetdev_callback\n");
#endif
	sbp = (struct sbp_softc *)xfer->sc;
	fw_xfer_free(xfer);
	sbp_bto(sbp);
	return;
}
#if 0
#else
static void sbp_reset_agent(struct sbp_softc *sbp, int i)
{
	struct fw_xfer *xfer;
	struct fw_pkt *fp;

#if SBP_DEBUG
	printf("sbp_reset_agent %d\n", i);
#endif
	xfer = fw_xfer_alloc();
	if(xfer == NULL){
		return;
	}
	xfer->send.len = 16;
	xfer->send.buf = malloc(xfer->send.len, M_DEVBUF, M_NOWAIT);
	if(xfer->send.buf == NULL){
		fw_xfer_free( xfer);
		return;
	}
	xfer->send.off = 0; 
	xfer->spd = 0;
	xfer->sc = (caddr_t)sbp;
	xfer->fc = sbp->fd.fc;
	xfer->act.hand = sbp_resetdev_callback;
	xfer->retry_req = fw_asybusy;

	fp = (struct fw_pkt *)xfer->send.buf;
	fp->mode.wreqq.dest_hi = htons(sbp->dev[i].login.cmd_hi);
	fp->mode.wreqq.dest_lo = htonl(sbp->dev[i].login.cmd_lo + 4);
	fp->mode.wreqq.tlrt = 0;
	fp->mode.wreqq.tcode = FWTCODE_WREQQ;
	fp->mode.wreqq.pri = 0;
	xfer->dst = 0xffc0 | sbp->dev[i].fwdev->dst;
	fp->mode.wreqq.dst = htons(xfer->dst);
	fp->mode.wreqq.data = htonl(0xf);
	fw_asyreq(xfer->fc, -1, xfer);
}
#endif
static void sbp_resetdev(struct sbp_softc *sbp)
{
	struct fw_xfer *xfer;
	struct fw_pkt *fp;
	int i;

#if SBP_DEBUG
	printf("sbp_resetdev\n");
#endif
	for( i = 0 ; i < SBP_NUM_LUNS ; i++){
		if((sbp->dev[i].fwdev != NULL) &&
		   (sbp->dev[i].fwdev->status == FWDEVATTACHED) &&
		   (sbp->dev[i].status == SBP_DEV_TOATTACH)){
			break;
		}
	}
	if(i < SBP_NUM_LUNS){
#if 0
		xfer = fw_xfer_alloc();
		if(xfer == NULL){
			return;
		}
		xfer->send.len = 16;
		xfer->send.buf = malloc(xfer->send.len, M_DEVBUF, M_NOWAIT);
		if(xfer->send.buf == NULL){
			fw_xfer_free( xfer);
			return;
		}
		xfer->send.off = 0; 
		xfer->spd = 0;
		xfer->sc = (caddr_t)sbp;
		xfer->fc = sbp->fd.fc;
		xfer->act.hand = sbp_resetdev_callback;
		xfer->retry_req = fw_asybusy;

		fp = (struct fw_pkt *)xfer->send.buf;
		fp->mode.wreqq.dest_hi = htons(sbp->dev[i].login.cmd_hi);
		fp->mode.wreqq.dest_lo = htonl(sbp->dev[i].login.cmd_lo + 4);
		fp->mode.wreqq.tlrt = 0;
		fp->mode.wreqq.tcode = FWTCODE_WREQQ;
		fp->mode.wreqq.pri = 0;
		xfer->dst = 0xffc0 | sbp->dev[i].fwdev->dst;
		fp->mode.wreqq.dst = htons(xfer->dst);
		fp->mode.wreqq.data = htonl(0xf);
		fw_asyreq(xfer->fc, -1, xfer);
#else
		sbp_reset_agent(sbp, i);
#endif
	}
}
static void sbp_login(struct sbp_softc *sbp)
{
	struct fw_xfer *xfer;
	struct fw_pkt *fp;
	struct sbp_ocb *ocb;
	int i, s;

#if SBP_DEBUG
	printf("sbp_login\n");
#endif
	for( i = 0 ; i < SBP_NUM_LUNS ; i++){
		if((sbp->dev[i].fwdev != NULL) &&
		   (sbp->dev[i].fwdev->status == FWDEVATTACHED) &&
		   (sbp->dev[i].status != SBP_DEV_ATTACHED)){
			break;
		}
	}
	if(i < SBP_NUM_LUNS){
#if 0
#else
		sbp_login_dev(sbp, i);
	}
}
static void sbp_login_dev(struct sbp_softc *sbp, int i)
{
	struct fw_xfer *xfer;
	struct fw_pkt *fp;
	struct sbp_ocb *ocb;
	int s;
#if SBP_DEBUG
	printf("sbp_login_dev %d\n", i);
#endif
#endif
		if ((ocb = sbp_get_ocb(sbp)) == NULL) {
			s = splcam();
			sbp->flags |= SBP_RESOURCE_SHORTAGE;
			splx(s);
			return;
		}
		ocb->flags = OCB_ACT_MGM;
		ocb->target = i;
		ocb->lun = getcsrdata(sbp->dev[i].fwdev, 0x14) & 0xffff;
#if SBP_DEBUG
		printf("Logical_Unit_Number: %x\n", ocb->lun);
#endif
		ocb->ccb = NULL;

		ocb->orb[0] = 0;
		ocb->orb[1] = 0;
		if(sbp->dev[i].status == SBP_DEV_RECONN){
#if SBP_DEBUG
			printf("status == RECONN %d\n", sbp->dev[i].status);
#endif
			ocb->orb[2] = 0;
			ocb->orb[3] = 0;
			ocb->orb[4] = htonl(ORB_NOTIFY | ORB_FUN_RCN | sbp->dev[i].login.id);
			ocb->orb[5] = 0;
			ocb->orb[6] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16) | 0x00000001);
			ocb->orb[7] = htonl((u_int32_t)&sbp->dev[i]);
		}else{
#if SBP_DEBUG
			printf("status != RECONN %d\n", sbp->dev[i].status);
#endif
			ocb->orb[2] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16));
			ocb->orb[3] = htonl(vtophys(&sbp->dev[i].login));
			ocb->orb[4] = htonl(ORB_NOTIFY | ORB_EXV | ocb->lun);
			ocb->orb[5] = htonl(sizeof(struct sbp_login_res));
			ocb->orb[6] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16) | 0x00000001);
			ocb->orb[7] = htonl((u_int32_t)&sbp->dev[i]);
		}

		xfer = fw_xfer_alloc();
		if(xfer == NULL){
			return;
		}
		xfer->send.len = 24;
		xfer->send.buf = malloc(xfer->send.len, M_DEVBUF, M_NOWAIT);
		if(xfer->send.buf == NULL){
			fw_xfer_free( xfer);
			return;
		}
		xfer->send.off = 0; 
		xfer->spd = 0;
		xfer->sc = (caddr_t)sbp;
		xfer->fc = sbp->fd.fc;
		xfer->act.hand = sbp_login_callback;
		xfer->retry_req = fw_asybusy;

		fp = (struct fw_pkt *)xfer->send.buf;
		fp->mode.wreqb.dest_hi = htons(sbp->dev[i].mgm_hi);
		fp->mode.wreqb.dest_lo = htonl(sbp->dev[i].mgm_lo);
		fp->mode.wreqb.tlrt = 0;
		fp->mode.wreqb.tcode = FWTCODE_WREQB;
		fp->mode.wreqb.pri = 0;
		xfer->dst = 0xffc0 | sbp->dev[i].fwdev->dst;
		fp->mode.wreqb.dst = htons(xfer->dst);
		fp->mode.wreqb.len = htons(8);
		fp->mode.wreqb.extcode = 0;
#if VMACCESS == 1
		fp->mode.wreqb.payload[0] = htonl(2);
		fp->mode.wreqb.payload[0] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16) | 2);
		fp->mode.wreqb.payload[1] = htonl(&ocb->orb[0]);
/*
printf("sbp %08x %08x %08x %08x\n", fp->mode.ld[0], fp->mode.ld[1], fp->mode.ld[2], fp->mode.ld[3]);
printf("sbp %08x %08x %08x %08x\n", fp->mode.ld[4], fp->mode.ld[5], fp->mode.ld[6], fp->mode.ld[7]);
*/
#else
		fp->mode.wreqb.payload[0] = 0;
		fp->mode.wreqb.payload[0] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16));
		fp->mode.wreqb.payload[1] = htonl(vtophys(&ocb->orb[0]));
#endif
		sbp_enqueue_ocb(sbp, i, ocb);

		fw_asyreq(xfer->fc, -1, xfer);
	}
#if 0
}
#endif
static void sbp_recv(struct fw_xfer *xfer){
	int target;
	struct fw_pkt *rfp, *sfp;
	struct sbp_softc *sbp;
	struct sbp_dev *dev;
	struct sbp_ocb *ocb;
	struct sbp_login_res *login_res = NULL;
	struct sbp_status *sbp_status;
	struct sbp_cmd_status *sbp_cmd_status;
/*
	u_int32_t *ld;
	ld = xfer->recv.buf;
printf("sbp %x %d %d %08x %08x %08x %08x\n",
			xfer->resp, xfer->recv.len, xfer->recv.off, ntohl(ld[0]), ntohl(ld[1]), ntohl(ld[2]), ntohl(ld[3]));
printf("sbp %08x %08x %08x %08x\n", ntohl(ld[4]), ntohl(ld[5]), ntohl(ld[6]), ntohl(ld[7]));
printf("sbp %08x %08x %08x %08x\n", ntohl(ld[8]), ntohl(ld[9]), ntohl(ld[10]), ntohl(ld[11]));
*/
	if(xfer->resp != 0){
		fw_xfer_free( xfer);
		return;
	}
	if(xfer->recv.buf == NULL){
		fw_xfer_free( xfer);
		return;
	}
	sbp = (struct sbp_softc *)xfer->sc;
	rfp = (struct fw_pkt *)xfer->recv.buf;
	if(rfp->mode.wreqb.tcode != FWTCODE_WREQB){
		fw_xfer_free( xfer);
		return;
	}
	sbp_status = (struct sbp_status *)rfp->mode.wreqb.payload;
	dev = (struct sbp_dev *)ntohl(rfp->mode.wreqb.dest_lo);
	target = SBP_DEV2LUN(sbp, dev);
#if SBP_DEBUG
printf("sbp stat src:%2x resp:%2x dead %2x len %3x stat %2x orb %8x:%8x\n",
	sbp_status->src, sbp_status->resp,sbp_status->dead, sbp_status->len, sbp_status->status, sbp_status->orb_hi, sbp_status->orb_lo );
printf("target: %d tlrt: %x\n", target, rfp->mode.wreqb.tlrt);
#endif
	if(target < 0 || target > SBP_NUM_LUNS){
		fw_xfer_free( xfer);
		return;
	}
	/* we have to reset the fetch agent if it's dead */
	if (sbp_status->dead)
		sbp_reset_agent(sbp, target);

	ocb = sbp_dequeue_ocb(sbp, target);
	if (ocb == NULL) {
		printf("No ocb on the queue for target %d.\n", target);
		fw_xfer_free( xfer);
		return;
	}
#define ORB_RES_CMPL 0
#define ORB_RES_FAIL 1
#define ORB_RES_ILLE 2
#define ORB_RES_VEND 3
	switch(ntohl(ocb->orb[4]) & ORB_FMT_MSK){
	case ORB_FMT_NOP:
		break;
	case ORB_FMT_VED:
		break;
	case ORB_FMT_STD:
		switch(ocb->flags & OCB_ACT_MASK){
		case OCB_ACT_MGM:
			switch(ntohl(ocb->orb[4]) & ORB_FUN_MSK){
			case ORB_FUN_LGI:
				if(sbp_status->resp != ORB_RES_CMPL){
					printf("%s:ORB err\n", device_get_nameunit(sbp->fd.dev));
					goto err;
				}
				login_res = &sbp->dev[target].login;
#if SBP_DEBUG
printf("%s: login :target %02x len %04x, ID %04x, cmd %08x%08x, recon_hold %04x\n", device_get_nameunit(sbp->fd.dev), target, ntohs(login_res->len), ntohs(login_res->id), ntohs(login_res->cmd_hi), ntohl(login_res->cmd_lo), ntohs(login_res->recon_hold));
#endif
				login_res->len = ntohs(login_res->len);
				login_res->id = ntohs(login_res->id);
				login_res->cmd_hi = ntohs(login_res->cmd_hi);
				login_res->cmd_lo = ntohl(login_res->cmd_lo);
				sbp->dev[target].status = SBP_DEV_TOATTACH;
				break;
			case ORB_FUN_RCN:
				login_res = &sbp->dev[target].login;
				if(sbp_status->resp != ORB_RES_CMPL ||
					sbp_status->status != 0){
					/* Reconnect failed */
					sbp->dev[target].status = SBP_DEV_RESET;
					login_res = NULL;
				}else{
					sbp->dev[target].status = SBP_DEV_ATTACHED;
#if SBP_DEBUG
printf("%s: reconnect :target %02x len %04x, ID %04x, cmd %08x%08x\n", device_get_nameunit(sbp->fd.dev), target, login_res->len, login_res->id, login_res->cmd_hi, login_res->cmd_lo);
#endif
				}
				break;
			default:
				break;
			}
			break;
		case OCB_ACT_CMD:
			if(ocb->ccb != NULL){
			struct scsi_sense_data *sense = &ocb->ccb->csio.sense_data;
/*
				u_int32_t *ld;
				ld = ocb->ccb->csio.data_ptr;
				if(ld != NULL && ocb->ccb->csio.dxfer_len != 0)
					printf("ptr %08x %08x %08x %08x\n", ld[0], ld[1], ld[2], ld[3]);
				else
					printf("ptr NULL\n");
printf("len %d\n", sbp_status->len);
*/
				untimeout(sbp_timeout, (caddr_t)ocb, ocb->ccb->ccb_h.timeout_ch);
				if(sbp_status->len > 1){
					sbp_cmd_status = (struct sbp_cmd_status *) sbp_status->data;
#if SBP_DEBUG
printf("status %d key %d code %d qlfr %d\n",
		sbp_cmd_status->status,
		sbp_cmd_status->s_key,
		sbp_cmd_status->s_code,
		sbp_cmd_status->s_qlfr);
#endif
					if(sbp_cmd_status->status == SCSI_STATUS_CHECK_COND || sbp_cmd_status->status == SCSI_STATUS_CMD_TERMINATED){
			if(sbp_cmd_status->sfmt == SBP_SFMT_CURR){
				sense->error_code = SSD_CURRENT_ERROR;
			}else{
				sense->error_code = SSD_DEFERRED_ERROR;
			}
			if(sbp_cmd_status->valid)
				sense->error_code |= SSD_ERRCODE_VALID;
			sense->flags = sbp_cmd_status->s_key;
			if(sbp_cmd_status->mark)
				sense->flags |= SSD_FILEMARK;
			if(sbp_cmd_status->eom)
				sense->flags |= SSD_EOM;
			if(sbp_cmd_status->ill_len)
				sense->flags |= SSD_ILI;
			sense->info[0] = ntohl(sbp_cmd_status->info) & 0xff;
			sense->info[1] =(ntohl(sbp_cmd_status->info) >> 8) & 0xff;
			sense->info[2] =(ntohl(sbp_cmd_status->info) >> 16) & 0xff;
			sense->info[3] =(ntohl(sbp_cmd_status->info) >> 24) & 0xff;
			sense->extra_len = (sbp_status->len - 5) * 4;
			sense->cmd_spec_info[0] = ntohl(sbp_cmd_status->cdb) & 0xff;
			sense->cmd_spec_info[1] = (ntohl(sbp_cmd_status->cdb) >> 8) & 0xff;
			sense->cmd_spec_info[2] = (ntohl(sbp_cmd_status->cdb) >> 16) & 0xff;
			sense->cmd_spec_info[3] = (ntohl(sbp_cmd_status->cdb) >> 24) & 0xff;
			sense->add_sense_code = sbp_cmd_status->s_code;
			sense->add_sense_code_qual = sbp_cmd_status->s_qlfr;
			sense->fru = sbp_cmd_status->fru;
			sense->sense_key_spec[0] = ntohl(sbp_cmd_status->s_keydep) & 0xff;
			sense->sense_key_spec[1] = (ntohl(sbp_cmd_status->s_keydep) >>8) & 0xff;
			sense->sense_key_spec[2] = (ntohl(sbp_cmd_status->s_keydep) >>16) & 0xff;

			ocb->ccb->csio.scsi_status = sbp_cmd_status->status;;
			ocb->ccb->ccb_h.status |= CAM_SCSI_STATUS_ERROR;
			ocb->ccb->ccb_h.status |= CAM_AUTOSNS_VALID;
/*
{
	u_int8_t j, *tmp;
	tmp = sense;
	for( j = 0 ; j < 32 ; j+=8){
		printf("sense %02x%02x %02x%02x %02x%02x %02x%02x\n", 
			tmp[j], tmp[j+1], tmp[j+2], tmp[j+3],
			tmp[j+4], tmp[j+5], tmp[j+6], tmp[j+7]);
	}

}
*/
			xpt_done(ocb->ccb);
					}
				}else{
					if(sbp_status->resp != ORB_RES_CMPL){
						ocb->ccb->ccb_h.status = CAM_REQ_INVALID;
						xpt_done(ocb->ccb);
					}else{
						ocb->ccb->ccb_h.status = CAM_REQ_CMP;
						xpt_done(ocb->ccb);
					}
				}
			}
			break;
		default:
			break;
		}
	}
/**/
err:
/* The received packet is usually small enough to be stored within
 * the buffer. In that case, the controller return ack_complete and
 * no respose is necessary.
 *
 * XXX fwohci.c and firewire.c should inform event_code such as 
 * ack_complete or ack_pending to upper driver.
 */
#if 0
	xfer->send.buf = malloc(12, M_DEVBUF, M_NOWAIT | M_ZERO);
	xfer->send.len = 12;
	xfer->send.off = 0;
	sfp = (struct fw_pkt *)xfer->send.buf;
	sfp->mode.wres.dst = rfp->mode.wreqb.src;
	xfer->dst = ntohs(sfp->mode.wres.dst);
	xfer->spd = 0;
	xfer->act.hand = sbp_loginres_callback;
	xfer->retry_req = fw_asybusy;

	sfp->mode.wres.tlrt = rfp->mode.wreqb.tlrt;
	sfp->mode.wres.tcode = FWTCODE_WRES;
	sfp->mode.wres.rtcode = 0;
	sfp->mode.wres.pri = 0;

	fw_asyreq(xfer->fc, -1, xfer);
#else
	fw_xfer_free(xfer);
#endif
/**/
	if(login_res != NULL){
		sbp_resetdev(sbp);
	} else if ((ntohl(ocb->orb[4]) & ORB_FUN_MSK) == ORB_FUN_RCN) {
		/* retry to login */
		sbp_login_dev(sbp, target);
	}
	sbp_free_ocb(sbp, ocb);
	return;
}
/*
 * sbp_attach()
 */
static int
sbp_attach(device_t dev)
{
	struct sbp_softc *sbp = ((struct sbp_softc *)device_get_softc(dev));
	struct cam_devq *devq;
	struct fw_xfer *xfer;
	int i, error;

#if SBP_DEBUG
	printf("sbp_attach\n");
#endif
#if 1
	error = bus_dma_tag_create(/*parent*/NULL, /*alignment*/1,
				/*boundary*/0,
				/*lowaddr*/BUS_SPACE_MAXADDR_32BIT,
				/*highaddr*/BUS_SPACE_MAXADDR,
				/*filter*/NULL, /*filterarg*/NULL,
				/*maxsize*/0x100000, /*nsegments*/0x20,
				/*maxsegsz*/0x8000,
				/*flags*/BUS_DMA_ALLOCNOW,
				&sbp->dmat);
	if (error != 0) {
		printf("sbp_attach: Could not allocate DMA tag "
			"- error %d\n", error);
			return (ENOMEM);
	}
#endif
#if 0
	devq = cam_simq_alloc(/*maxopenings*/1);
#else
	devq = cam_simq_alloc(/*maxopenings*/SBP_NUM_LUNS);
#endif
	if (devq == NULL)
		return (ENXIO);

	for( i = 0 ; i < SBP_NUM_LUNS ; i++){
		sbp->dev[i].fwdev = NULL;
		sbp->dev[i].status = SBP_DEV_RESET;
		sbp->dev[i].login.id = 0;
		sbp->dev[i].login.cmd_hi = 0;
		sbp->dev[i].login.cmd_lo = 0;
	}

	sbp->sim = cam_sim_alloc(sbp_action, sbp_poll, "sbp", sbp,
				 device_get_unit(dev),
				 /*untagged*/1, /*tagged*/0, devq);

	if (sbp->sim == NULL) {
		cam_simq_free(devq);
		return (ENXIO);
	}

	sbp->free_ocbs = NULL;
	for (i = 255; i >= 0; i--) {
		sbp->ocb[i].tag = i;
		sbp_free_ocb(sbp, &sbp->ocb[i]);
	}

	sbp->dummy_ocb = &sbp->ocb[256];

	sbp->dummy_ocb->orb[0] = htonl(1 << 31);
	sbp->dummy_ocb->orb[1] = 0;
	sbp->dummy_ocb->orb[2] = 0;
	sbp->dummy_ocb->orb[3] = 0;
	sbp->dummy_ocb->orb[4] = htonl(ORB_NOTIFY | ORB_FMT_NOP);
	sbp->dummy_ocb->orb[5] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16) | 1);
	sbp->dummy_ocb->orb[6] = htonl((u_int32_t)sbp->dummy_ocb->orb);

	if (xpt_bus_register(sbp->sim, /*bus*/0) != CAM_SUCCESS) {
		cam_sim_free(sbp->sim, /*free_devq*/TRUE);
		return (ENXIO);
	}

	xfer = fw_xfer_alloc();
	xfer->act.hand = sbp_recv;
	xfer->act_type = FWACT_XFER;
	xfer->fc = sbp->fd.fc;
	xfer->sc = (caddr_t)sbp;

	sbp->fwb.start_hi = 0x1;
	sbp->fwb.start_lo = (u_int32_t)sbp->dev;
	sbp->fwb.addrlen = SBP_NUM_LUNS * sizeof(struct sbp_dev);
	sbp->fwb.xfer = xfer;
	fw_bindadd(sbp->fd.fc, &sbp->fwb);

	sbp_cam_rescan(sbp);	/* have CAM rescan the bus */

	return (0);
}
static int
sbp_detach(device_t dev)
{
	/* XXX */ 
	struct sbp_softc *sbp = ((struct sbp_softc *)device_get_softc(dev));

	xpt_bus_deregister(cam_sim_path(sbp->sim));
	return (0);
}

static void
sbp_cam_rescan_callback(struct cam_periph *periph, union ccb *ccb)
{
	xpt_free_path(ccb->ccb_h.path);
	free(ccb, M_TEMP);
}

static void
sbp_cam_rescan(struct sbp_softc *sbp)
{
	struct cam_path *path;
	union ccb *ccb = malloc(sizeof(union ccb), M_TEMP, M_WAITOK);

	memset(ccb, 0, sizeof(union ccb));

	if (xpt_create_path(&path, xpt_periph, cam_sim_path(sbp->sim),
			    CAM_TARGET_WILDCARD, CAM_LUN_WILDCARD)
	    != CAM_REQ_CMP)
		return;

	xpt_setup_ccb(&ccb->ccb_h, path, 5/*priority (low)*/);
	ccb->ccb_h.func_code = XPT_SCAN_BUS;
	ccb->ccb_h.cbfcnp = sbp_cam_rescan_callback;
	ccb->crcn.flags = CAM_FLAG_NONE;
	xpt_action(ccb);

	/* The scan is in progress now. */
}
static void
sbp_timeout(void *arg)
{
	struct sbp_ocb *ocb = (struct sbp_ocb *)arg;
	union ccb *ccb = ocb->ccb;
	struct sbp_softc *sbp = (struct sbp_softc *)ccb->ccb_h.ccb_sbp_ptr;

	ccb->ccb_h.status = CAM_TID_INVALID;
	xpt_done(ccb);
	return;
}
static void
sbp_action(struct cam_sim *sim, union ccb *ccb)
{

	struct sbp_softc *sbp = (struct sbp_softc *)sim->softc;

	switch (ccb->ccb_h.func_code) {
	case XPT_SCSI_IO:
	case XPT_RESET_DEV:
	case XPT_GET_TRAN_SETTINGS:
	case XPT_SET_TRAN_SETTINGS:
	case XPT_CALC_GEOMETRY:
		if (sbp == NULL) {
			printf("sbp:%d:%d:func_code 0x%04x: "
				"Invalid target (target needed)\n",
				ccb->ccb_h.target_id, ccb->ccb_h.target_lun,
				ccb->ccb_h.func_code);

			ccb->ccb_h.status = CAM_TID_INVALID;
			xpt_done(ccb);
			return;
		}
		break;
	case XPT_PATH_INQ:
	case XPT_NOOP:
		/* The opcodes sometimes aimed at a target (sc is valid),
		 * sometimes aimed at the SIM (sc is invalid and target is
		 * CAM_TARGET_WILDCARD)
		 */
		if (sbp == NULL && ccb->ccb_h.target_id != CAM_TARGET_WILDCARD) {
#if SBP_DEBUG > 0
			printf("sbp:%d:%d:func_code 0x%04x: "
				"Invalid target (no wildcard)\n",
				ccb->ccb_h.target_id, ccb->ccb_h.target_lun,
				ccb->ccb_h.func_code);
#endif
			ccb->ccb_h.status = CAM_TID_INVALID;
			xpt_done(ccb);
			return;
		}
		break;
	default:
		/* XXX Hm, we should check the input parameters */
	}

	switch (ccb->ccb_h.func_code) {
	case XPT_SCSI_IO:
	{
		struct ccb_scsiio *csio;
		struct sbp_ocb *ocb;
		struct fw_xfer *xfer;
		struct fw_pkt *fp;
		u_int32_t i, size;
		u_int32_t *ld;
		int s;

		csio = &ccb->csio;

#if SBP_DEBUG > 0
		printf("%s:%d:%d:%d:XPT_SCSI_IO: "
			"cmd: 0x%02x, flags: 0x%02x, "
			"%db cmd/%db data/%db sense\n",
			device_get_nameunit(sbp->fd.dev), cam_sim_path(sbp->sim),
			ccb->ccb_h.target_id, ccb->ccb_h.target_lun,
			csio->cdb_io.cdb_bytes[0],
			ccb->ccb_h.flags & CAM_DIR_MASK,
			csio->cdb_len, csio->dxfer_len,
			csio->sense_len);
#endif
		if(ccb->ccb_h.target_id < 0 ||
			sbp->dev[ccb->ccb_h.target_id].fwdev == NULL ||
			sbp->dev[ccb->ccb_h.target_id].fwdev->status != FWDEVATTACHED){
			ccb->ccb_h.status = CAM_REQ_INVALID;
			xpt_done(ccb);
			return;
		}
		/* XXX allow only 1 unit */
		if (ccb->ccb_h.target_lun != 0) {
			ccb->ccb_h.status = CAM_REQ_INVALID;
			xpt_done(ccb);
			return;
		}
		if ((ocb = sbp_get_ocb(sbp)) == NULL) {
			s = splcam();
			sbp->flags |= SBP_RESOURCE_SHORTAGE;
			splx(s);
			return;
		}
		ocb->flags = OCB_ACT_CMD;
		ocb->target = ccb->ccb_h.target_id;
		ocb->lun = ccb->ccb_h.target_lun;
		ocb->ccb = ccb;
#if 0
		ccb->ccb_h.ccb_ocb_ptr = ocb;
#endif
		ccb->ccb_h.ccb_sbp_ptr = sbp;
		ocb->orb[0] = htonl(1 << 31);
		ocb->orb[1] = 0;
#if 0
#define SBP_ORB_MAXP (1 << 16)
		if(ccb->csio.dxfer_len < SBP_ORB_MAXP){
#if VMACCESS == 1
			ocb->orb[2] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16) | 2);
			ocb->orb[3] = htonl(ccb->csio.data_ptr);
#else
			ocb->orb[2] = htonl(0);
			ocb->orb[2] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16) );
			ocb->orb[3] = (ccb->csio.data_ptr != NULL || csio->dxfer_len > 0) ?
				htonl(vtophys(ccb->csio.data_ptr)) : NULL;
#endif
/*
			ld = ccb->csio.data_ptr;
			if(ld != NULL && ccb->csio.dxfer_len != 0)
				printf("ptr %08x %08x %08x %08x\n", ld[0], ld[1], ld[2], ld[3]);
*/
			ocb->orb[4] = htonl(ORB_NOTIFY | ORB_CMD_SPD(2) | ORB_CMD_MAXP(7) | csio->dxfer_len);
		}else{
#if VMACCESS == 1
			ocb->orb[2] = htonl(2);
			ocb->orb[2] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16) | 2);
			ocb->orb[3] = htonl(ocb->ind_ptr);
#else
			ocb->orb[2] = htonl(0);
			ocb->orb[2] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16) );
			ocb->orb[3] = htonl(vtophys(ocb->ind_ptr));
#endif
			size = csio->dxfer_len;
			for( i = 0 ; i < 0x20 ; i++){
#if VMACCESS == 1
				ocb->ind_ptr[i].hi = 2;
				ocb->ind_ptr[i].lo
		= htonl((((u_int32_t)ccb->csio.data_ptr)) + (i << 15));
#else
				ocb->ind_ptr[i].hi = 0;
				ocb->ind_ptr[i].lo
		= htonl((vtophys(ccb->csio.data_ptr)) + (i << 15));
#endif
				if(size <= (1 << 15)){
					ocb->ind_ptr[i].hi |= htonl(size << 16);
					break;
				}
				ocb->ind_ptr[i].hi |= htonl(1 << 31);
				size -= (1 << 15);
			}
			ocb->orb[4] = htonl(ORB_NOTIFY | ORB_CMD_SPD(2) | ORB_CMD_MAXP(7) | ORB_CMD_PTBL | (i + 1));
		}
#else
		ocb->orb[2] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16) );
		ocb->orb[3] = htonl(vtophys(ocb->ind_ptr));
		/* S400 2^(9+2)=2048bytes */
		ocb->orb[4] = htonl(ORB_NOTIFY | ORB_CMD_SPD(2) | ORB_CMD_MAXP(9));
#endif
		if((ccb->ccb_h.flags & CAM_DIR_MASK) == CAM_DIR_IN){
			ocb->orb[4] |= htonl(ORB_CMD_IN);
		}
		if (csio->ccb_h.flags & CAM_CDB_POINTER) {
			bcopy(csio->cdb_io.cdb_ptr, &ocb->orb[5], (csio->cdb_len + 3) & ~0x3);
		} else {
			bcopy(&csio->cdb_io.cdb_bytes, &ocb->orb[5], (csio->cdb_len + 3) &~0x3);
		}
/*
printf("ORB %08x %08x %08x %08x\n", ntohl(ocb->orb[0]), ntohl(ocb->orb[1]), ntohl(ocb->orb[2]), ntohl(ocb->orb[3]));
printf("ORB %08x %08x %08x %08x\n", ntohl(ocb->orb[4]), ntohl(ocb->orb[5]), ntohl(ocb->orb[6]), ntohl(ocb->orb[7]));
*/
		xfer = fw_xfer_alloc();
#if 1
		ccb->ccb_h.ccb_xfer_ptr = xfer;
#endif
		if(xfer == NULL){
			return;
		}
		xfer->send.len = 24;
		xfer->send.buf = malloc(xfer->send.len, M_DEVBUF, M_NOWAIT);
		if(xfer->send.buf == NULL){
			fw_xfer_free( xfer);
			return;
		}
		xfer->send.off = 0; 
		xfer->spd = 0;
		xfer->sc = (caddr_t)sbp;
		xfer->fc = sbp->fd.fc;
		xfer->act.hand = sbp_login_callback;
		xfer->retry_req = fw_asybusy;
		fp = (struct fw_pkt *)xfer->send.buf;
		fp->mode.wreqb.dest_hi = htons(sbp->dev[ocb->target].login.cmd_hi);
		fp->mode.wreqb.tlrt = 0;
		fp->mode.wreqb.dest_lo = htonl(sbp->dev[ocb->target].login.cmd_lo + 8);
		fp->mode.wreqb.tcode = FWTCODE_WREQB;
		fp->mode.wreqb.pri = 0;
		xfer->dst = 0xffc0 | sbp->dev[ocb->target].fwdev->dst;
		fp->mode.wreqb.dst = htons(xfer->dst);
		fp->mode.wreqb.len = htons(8);
		fp->mode.wreqb.extcode = 0;
#if VMACCESS == 1
		fp->mode.wreqb.payload[0] = htonl(2);
		fp->mode.wreqb.payload[0] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16)|2 );
		fp->mode.wreqb.payload[1] = htonl(&ocb->orb[0]);
#else
		fp->mode.wreqb.payload[0] = 0;
		fp->mode.wreqb.payload[0] = htonl(((sbp->fd.fc->nodeid | 0xffc0 )<< 16));
		fp->mode.wreqb.payload[1] = htonl(vtophys(&ocb->orb[0]));
#endif
#if 0
		sbp_enqueue_ocb(sbp, ccb->ccb_h.target_id, ocb);

		if(fw_asyreq(xfer->fc, -1, xfer) != 0){
			ccb->ccb_h.status = CAM_REQ_INVALID;
			xpt_done(ccb);
		}
		ccb->ccb_h.timeout_ch = timeout(sbp_timeout, (caddr_t)ocb,
			(ccb->ccb_h.timeout * hz) / 1000);
#else
		if (ccb->csio.dxfer_len > 0) {
			bus_dmamap_load(/*dma tag*/sbp->dmat,
					/*dma map*/NULL,
					ccb->csio.data_ptr,
					ccb->csio.dxfer_len,
					sbp_execute_ocb,
					ocb,
					/*flags*/0);
		} else
			sbp_execute_ocb(ocb, NULL, 0, 0);
#endif
		break;
	}
	case XPT_CALC_GEOMETRY:
	{
		struct ccb_calc_geometry *ccg;
		u_int32_t size_mb;
		u_int32_t secs_per_cylinder;
		int extended = 0;
		ccg = &ccb->ccg;
#if SBP_DEBUG > 0
		printf("%s:%d:%d:%d:XPT_CALC_GEOMETRY: "
			"Volume size = %d\n",
			device_get_nameunit(sbp->fd.dev), cam_sim_path(sbp->sim),
			ccb->ccb_h.target_id, ccb->ccb_h.target_lun,
			ccg->volume_size);
#endif

		size_mb = ccg->volume_size
			/ ((1024L * 1024L) / ccg->block_size);

		if (size_mb >= 1024 && extended) {
			ccg->heads = 255;
			ccg->secs_per_track = 63;
		} else {
			ccg->heads = 64;
			ccg->secs_per_track = 32;
		}
		secs_per_cylinder = ccg->heads * ccg->secs_per_track;
		ccg->cylinders = ccg->volume_size / secs_per_cylinder;
		ccb->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;
	}
	case XPT_RESET_BUS:		/* Reset the specified SCSI bus */
	{

#if SBP_DEBUG > 0
		printf("%s:%d:XPT_RESET_BUS: \n",
			device_get_nameunit(sbp->fd.dev), cam_sim_path(sbp->sim));
#endif

		ccb->ccb_h.status = CAM_REQ_INVALID;
		xpt_done(ccb);
		break;
	}
	case XPT_PATH_INQ:		/* Path routing inquiry */
	{
		struct ccb_pathinq *cpi = &ccb->cpi;
		
#if SBP_DEBUG > 0
		printf("%s:%d:%d:%d:XPT_PATH_INQ:.\n",
			device_get_nameunit(sbp->fd.dev),
			cam_sim_path(sbp->sim),
			ccb->ccb_h.target_id, ccb->ccb_h.target_lun);
#endif
		cpi->version_num = 1; /* XXX??? */
		cpi->hba_inquiry = 0;
		cpi->target_sprt = 0;
		cpi->hba_misc = 0;
		cpi->hba_eng_cnt = 0;
		cpi->max_target = 7;
		cpi->max_lun = SBP_NUM_LUNS -1;
		cpi->initiator_id = SBP_INITIATOR;
		cpi->bus_id = sim->bus_id;
		cpi->base_transfer_speed = 93;
		strncpy(cpi->sim_vid, "FreeBSD", SIM_IDLEN);
		strncpy(cpi->hba_vid, "SBP", HBA_IDLEN);
		strncpy(cpi->dev_name, sim->sim_name, DEV_IDLEN);
		cpi->unit_number = sim->unit_number;

		cpi->ccb_h.status = CAM_REQ_CMP;
		xpt_done(ccb);
		break;
	}
	default:
		ccb->ccb_h.status = CAM_REQ_INVALID;
		xpt_done(ccb);
		break;
	}
	return;
}
#if 1
static void
sbp_execute_ocb(void *arg,  bus_dma_segment_t *segments, int seg, int error)
{
	int i;
	struct sbp_ocb *ocb;
	union ccb *ccb;
	struct fw_xfer *xfer;

	ocb = (struct sbp_ocb *)arg;
	if (seg == 1) {
		ocb->orb[3] = htonl(segments[0].ds_addr);
		ocb->orb[4] |= htonl(segments[0].ds_len);
	} else if(seg > 1) {
#if SBP_DEBUG
		printf("sbp_execute_ocb: seg %d", seg);
		for (i = 0; i < seg; i++)
			printf(", %X:%d", segments[i].ds_addr,
						segments[i].ds_len);
		printf("\n");
#endif
		for (i = 0; i < seg; i++) {
			ocb->ind_ptr[i].hi = htonl(segments[i].ds_len << 16);
			ocb->ind_ptr[i].lo = htonl(segments[i].ds_addr);
		}
		ocb->orb[4] |= htonl(ORB_CMD_PTBL | seg);
	}
	
	ccb = ocb->ccb;
	xfer = ccb->ccb_h.ccb_xfer_ptr;
	sbp_enqueue_ocb(ccb->ccb_h.ccb_sbp_ptr, ccb->ccb_h.target_id, ocb);

	if(fw_asyreq(xfer->fc, -1, xfer) != 0){
			ccb->ccb_h.status = CAM_REQ_INVALID;
			xpt_done(ccb);
	}
	ccb->ccb_h.timeout_ch = timeout(sbp_timeout, (caddr_t)ocb,
			(ccb->ccb_h.timeout * hz) / 1000);
}
#endif

static void
sbp_poll(struct cam_sim *sim)
{       
	return;
}

static struct sbp_ocb *
sbp_dequeue_ocb(struct sbp_softc *sbp, int i)
{
	struct sbp_ocb *ocb;
	int s = splcam();
	if ((ocb = sbp->dev[i].ocbs) != NULL)
		sbp->dev[i].ocbs = sbp->dev[i].ocbs->ocb;
	splx(s);
	return (ocb);
}
static void
sbp_enqueue_ocb(struct sbp_softc *sbp, int i, struct sbp_ocb *ocb)
{
	int s = splcam();
	ocb->ocb = sbp->dev[i].ocbs;
	sbp->dev[i].ocbs = ocb;
	splx(s);
}
static struct sbp_ocb *
sbp_get_ocb(struct sbp_softc *sbp)
{
	struct sbp_ocb *ocb;
	int s = splcam();
	if ((ocb = sbp->free_ocbs) != NULL)
		sbp->free_ocbs = sbp->free_ocbs->ocb;
	splx(s);
	return (ocb);
}

static void
sbp_free_ocb(struct sbp_softc *sbp, struct sbp_ocb *ocb)
{
	int s = splcam();
	if ((sbp->flags & SBP_RESOURCE_SHORTAGE) != 0 &&
	    (ocb->ccb->ccb_h.status & CAM_RELEASE_SIMQ) == 0) {
		ocb->ccb->ccb_h.status |= CAM_RELEASE_SIMQ;
		sbp->flags &= ~SBP_RESOURCE_SHORTAGE;
	}
	ocb->flags = 0;
	ocb->ocb = sbp->free_ocbs;
	sbp->free_ocbs = ocb;
	splx(s);
}

static devclass_t sbp_devclass;

static device_method_t sbp_methods[] = {
	/* device interface */
	DEVMETHOD(device_identify,	sbp_identify),
	DEVMETHOD(device_probe,		sbp_probe),
	DEVMETHOD(device_attach,	sbp_attach),
	DEVMETHOD(device_detach,	sbp_detach),

	{ 0, 0 }
};

static driver_t sbp_driver = {
	"sbp",
	sbp_methods,
	sizeof(struct sbp_softc),
};
DRIVER_MODULE(sbp, firewire, sbp_driver, sbp_devclass, 0, 0);
